#include <stdio.h>
#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include "pad.h"

/*

Second example based on the tutorial from Henry S Fortuna...

A Breakout clone that only uses multi-colored boxes.
Move the paddle sideways witht he pad, press Start+Select to end game.

Read the six first lessons from Henry S Fortuna. (http://www.hsfortuna.pwp.blueyonder.co.uk/)
The last lesson you read should be the PS2 Z-Buffer lesson.

A few assumptions...
We know how to write the GIF Tags/Regs by now so I'm going to cut down on the comments about the GIF stuff.
We are going to allocate 4096 bytes for the GIF Tags/Regs, this is done with the sps2Allocate to get physically contiguous memory.
Memory allocated with the Linux functions can reside in any give memory position and may not be contigous.
The operating system will take care of the memory functions for you, however for SPS2 we need physically contiguous memory.
Every box occupies 3 GIF Regs and we need one GIF Tag which gives us (4096-16)/(16*3) = 85 boxes maximum.
With 5 boxes for the outline pad and ball, that leaves us with 80 bricks.
The 80 bricks will be divied into four rows with 20 bricks each in different colors.
All visible bricks and the outline will have a Z value of 200.
The background color will be black and if we remove a brick we just set this brick to black with a Z value of 100.
The ball and paddle will be drawn with a Z value of 300, this will guarantee that these are always on top of everything else.
Remember that everything with a higher Z value will be drawn on top of anything with a lower Z value.
To keep it simple we are going to plot all 85 boxes per frame.
This is done by filling the allocated memory with one GIF Tag and the GIF Regs for the 85 boxes.
Then the memory is written straight to the GIF with the help of the DMAC over DMA channel 2.

Once again.
If you have been programming on the PC (Intel or compatible), you will probably think a word is 16 bits...
Have a look at page 5 in EEUSER_E.PDF, a word is 32 bits and a QuadWord is 32*4=128 bits !!!

Written by Bamse 2004...
bamse@playstation2-linux.com

*/

#define PAD_MAX 570
#define PAD_MIN 10

// Function prototypes
void AddGIFTag(char *& pMem);
void AddBox(char *& pMem, int iX, int iY, int iWidth, int iHeight, int iRed, int iGreen, int iBlue, int iZvalue);
int CountBricks(void);
void ResetBricks(void);

// Globals
int iPaddlePos = 290;		// Paddle in the middle
int iBallXPos = 318;		// Ball initialization.
int iBallYPos = 200;
int iBallXDir = -2;
int iBallYDir = 2;
int iSPS2Descriptor;		// SPS2 handle
sps2Memory_t * pMemory;		// Info struct about allocated memory
char * pMemPosition;		// The memory position
int iBricks[4][20];			// The rows of bricks.
Dn_CHCR_t chcr;	  			// Channel control register used to control the DMA transfer.


int main(void) 
{
	// Initialise the Pad connected to port 1 of the PS2, we initilize it as analogue so we can read the pressure.
	if(pad_init (PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE ) == 0)
	{
		printf("Unable to initialise the pads\n");
		pad_cleanup (PAD_0);
		exit(0);
	}
  	
  	// Initialize sps2     
  	iSPS2Descriptor = sps2Init();
  	if(iSPS2Descriptor < 0)	{
		printf("sps2Init() failed\n");
  		exit(0);
  	}

	// Make all bricks visible.
	ResetBricks();

  	// Allocate 4k of uncached SPS2 memory.
	// We are using uncached memory which is slower byt we don't have to worry
	// about flushing the cache before the DMA transfer.
  	pMemory=sps2Allocate(4096, SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED, iSPS2Descriptor);

  	// Initialise the screen.  
  	sps2UScreenInit(0);

  	do 
  	{

		// Read the pad.
		pad_update (PAD_0);

		// Update paddle posistion for right direction
		// We are doing this three times, that means that the harder you press the faster the paddle go...
		// Slight pressure
		if(pad[0].pressures[PAD_PLEFT] > 0.3) {
			if (iPaddlePos>PAD_MIN) {
				iPaddlePos--;
			}
		}
		// Medium pressure
		if(pad[0].pressures[PAD_PLEFT] > 0.6) {
			if (iPaddlePos>PAD_MIN) {
				iPaddlePos--;
			}
		}
		// Hard pressure pressure
		if(pad[0].pressures[PAD_PLEFT] > 0.9) {
			if (iPaddlePos>PAD_MIN) {
				iPaddlePos--;
			}
		}

		// Update paddle posistion for right direction
		// We are doing this three times, that means that the harder you press the faster the paddle go...
		// Slight pressure
		if(pad[0].pressures[PAD_PRIGHT] > 0.3) {
			if (iPaddlePos<PAD_MAX) {
				iPaddlePos++;
			}
		}
		// Medium pressure
		if(pad[0].pressures[PAD_PRIGHT] > 0.6) {
			if (iPaddlePos<PAD_MAX) {
				iPaddlePos++;
			}
		}
		// Hard pressure pressure
		if(pad[0].pressures[PAD_PRIGHT] > 0.9) {
			if (iPaddlePos<PAD_MAX) {
				iPaddlePos++;
			}
		}

		// Update the ball position...
		iBallXPos += iBallXDir;
		iBallYPos += iBallYDir;

		// Did we bounce against the right wall???
		if (iBallXPos >= 626) {
			iBallXDir = -iBallXDir;
		}

		// Did we bounce against the left wall???
		if (iBallXPos <= 10) {
			iBallXDir = -iBallXDir;
		}

		// Did we bounce against the upper wall???
		if (iBallYPos <= 10) {
			iBallYDir = -iBallYDir;
		}

		// Did we bounce against the PADDLE???
		if (iBallYPos >= 466) {
			// Where is the paddle compared to the ball???
			int iDiff = iBallXPos - iPaddlePos;
			// Missed the paddle ???
			if (iDiff < -4 || iDiff > 60) {
				// Reset ball since we missed the paddle...
				iBallYPos = 200;
				iBallXPos = 318;
				iBallXDir = -2;
				iBallYDir = 2;
			} else {
				// We hit the paddle, the ball bounced...
				// Let's calculate where on the paddle we hit and set different directions.
				// The speed will be kept consistent since iBallYDir+iBallXDir will always be 4.
				// Did we hit far Left end ???
				if (iDiff <= 8) {
					iBallYDir = -1;
					iBallXDir = -3;
				// Mid Left paddle ???
				} else if (iDiff > 8 && iDiff <= 18) {
					iBallYDir = -2;
					iBallXDir = -2;
				// Center Left paddle ???
				} else if (iDiff > 18 && iDiff <= 28) {
					iBallYDir = -3;
					iBallXDir = -1;
				// Center Right paddle ???
				} else if (iDiff > 28 && iDiff <= 38) {
					iBallYDir = -3;
					iBallXDir = 1;
				// Mid Right paddle ???
				} else if (iDiff > 38 && iDiff <= 48) {
					iBallYDir = -2;
					iBallXDir = 2;
				// Far Right paddle ???
				} else if (iDiff > 48) {
					iBallYDir = -1;
					iBallXDir = 3;
				}
			}
		}

		// Did we bounce against the first (upper) Row ???
		// Going up, Y value between 50 and 40...
		if (iBallYDir<0 && (iBallYPos <= 50 && iBallYPos > 40 )) {
			// Where is the paddle compared to the bricks???
			int iDiff = (iBallXPos - 20)/30;
			// We hit a brick.
			if (iDiff>-1 && iDiff<20) {
				// Is the brick visible ?
				if (iBricks[0][iDiff]) {
					iBricks[0][iDiff] = 0;
					iBallYDir = -iBallYDir;
				}
			}
		}
		// Going down, Y value between 46 and 36...
		// Why 46 and 36 ??? The ball is 4x4 pixels...
		if (iBallYDir>0 && (iBallYPos < 46 && iBallYPos >= 36 )) {
			// Where is the paddle compared to the bricks???
			int iDiff = (iBallXPos - 20)/30;
			// We hit a brick.
			if (iDiff>-1 && iDiff<20) {
				// Is the brick visible ?
				if (iBricks[0][iDiff]) {
					iBricks[0][iDiff] = 0;
					iBallYDir = -iBallYDir;
				}
			}
		}

		// Did we bounce against the second Row ???
		// Going up, Y value between 90 and 80...
		if (iBallYDir<0 && (iBallYPos <= 90 && iBallYPos > 80 )) {
			// Where is the paddle compared to the bricks???
			int iDiff = (iBallXPos - 20)/30;
			// We hit a brick.
			if (iDiff>-1 && iDiff<20) {
				// Is the brick visible ?
				if (iBricks[1][iDiff]) {
					iBricks[1][iDiff] = 0;
					iBallYDir = -iBallYDir;
				}
			}
		}
		// Going down, Y value between 86 and 76...
		if (iBallYDir>0 && (iBallYPos < 86 && iBallYPos >= 76 )) {
			// Where is the paddle compared to the bricks???
			int iDiff = (iBallXPos - 20)/30;
			// We hit a brick.
			if (iDiff>-1 && iDiff<20) {
				// Is the brick visible ?
				if (iBricks[1][iDiff]) {
					iBricks[1][iDiff] = 0;
					iBallYDir = -iBallYDir;
				}
			}
		}

		// Did we bounce against the third Row ???
		// Going up, Y value between 130 and 120...
		if (iBallYDir<0 && (iBallYPos <= 130 && iBallYPos > 120 )) {
			// Where is the paddle compared to the bricks???
			int iDiff = (iBallXPos - 20)/30;
			// We hit a brick.
			if (iDiff>-1 && iDiff<20) {
				// Is the brick visible ?
				if (iBricks[2][iDiff]) {
					iBricks[2][iDiff] = 0;
					iBallYDir = -iBallYDir;
				}
			}
		}
		// Going down, Y value between 126 and 116...
		if (iBallYDir>0 && (iBallYPos < 126 && iBallYPos >= 116 )) {
			// Where is the paddle compared to the bricks???
			int iDiff = (iBallXPos - 20)/30;
			// We hit a brick.
			if (iDiff>-1 && iDiff<20) {
				// Is the brick visible ?
				if (iBricks[2][iDiff]) {
					iBricks[2][iDiff] = 0;
					iBallYDir = -iBallYDir;
				}
			}
		}

		// Did we bounce against the fourt Row ???
		// Going up, Y value between 170 and 160...
		if (iBallYDir<0 && (iBallYPos <= 170 && iBallYPos > 160 )) {
			// Where is the paddle compared to the bricks???
			int iDiff = (iBallXPos - 20)/30;
			// We hit a brick.
			if (iDiff>-1 && iDiff<20) {
				// Is the brick visible ?
				if (iBricks[3][iDiff]) {
					iBricks[3][iDiff] = 0;
					iBallYDir = -iBallYDir;
				}
			}
		}
		// Going down, Y value between 166 and 154...
		if (iBallYDir>0 && (iBallYPos < 166 && iBallYPos >= 154 )) {
			// Where is the paddle compared to the bricks???
			int iDiff = (iBallXPos - 20)/30;
			// We hit a brick.
			if (iDiff>-1 && iDiff<20) {
				// Is the brick visible ?
				if (iBricks[3][iDiff]) {
					iBricks[3][iDiff] = 0;
					iBallYDir = -iBallYDir;
				}
			}
		}

		// See if all bricks have been hit.
		if (CountBricks()==0) {
			// Just reset the bricks.
			ResetBricks();
		}

    	// Clear the screen buffer to (r,g,b)
    	sps2UScreenClear(0x00,0x00,0x00);

		// Get first memory position from the allocated memory.
		pMemPosition = (char *) pMemory->pvStart;

		// Add the GIF tag to the begining of memory.
		AddGIFTag(pMemPosition);

    	// Upper box.
    	AddBox(pMemPosition, 0, 0, 640, 10, 0xFF, 0xFF, 0xFF, 200);

    	// Right box.
    	AddBox(pMemPosition, 630, 0, 10, 480, 0xFF, 0xFF, 0xFF, 200);

    	// Left box.
    	AddBox(pMemPosition, 0, 0, 10, 480, 0xFF, 0xFF, 0xFF, 200);

		// Row 1, Red bricks...
		for (int iCol = 0; iCol<20 ; iCol++) {
			// Is the brick visible ???
			if (iBricks[0][iCol]) {
				AddBox(pMemPosition, 20 + iCol*30, 40, 28, 10, 0xFF, 0x00, 0x00, 200);
			// Otherwise we'll make it black.
			} else {
				AddBox(pMemPosition, 20 + iCol*30, 40, 28, 10, 0x00, 0x00, 0x00, 100);
			}
		}		

		// Row 2, Yellow bricks...
		for (int iCol = 0; iCol<20 ; iCol++) {
			// Is the brick visible ???
			if (iBricks[1][iCol]) {
				AddBox(pMemPosition, 20 + iCol*30, 80, 28, 10, 0xFF, 0xFF, 0x00, 200);
			// Otherwise we'll make it black.
			} else {
				AddBox(pMemPosition, 20 + iCol*30, 80, 28, 10, 0x00, 0x00, 0x00, 100);
			}
		}		

		// Row 3, Green bricks...
		for (int iCol = 0; iCol<20 ; iCol++) {
			// Is the brick visible ???
			if (iBricks[2][iCol]) {
				AddBox(pMemPosition, 20 + iCol*30, 120, 28, 10, 0x00, 0xFF, 0x00, 200);
			// Otherwise we'll make it black.
			} else {
				AddBox(pMemPosition, 20 + iCol*30, 120, 28, 10, 0x00, 0x00, 0x00, 100);
			}
		}		

		// Row 4, Blue bricks...
		for (int iCol = 0; iCol<20 ; iCol++) {
			// Is the brick visible ???
			if (iBricks[3][iCol]) {
				AddBox(pMemPosition, 20 + iCol*30, 160, 28, 10, 0x00, 0x00, 0xFF, 200);
			// Otherwise we'll make it black.
			} else {
				AddBox(pMemPosition, 20 + iCol*30, 160, 28, 10, 0x00, 0x00, 0x00, 100);
			}
		}		

	   	// Paddle box.
    	AddBox(pMemPosition, iPaddlePos, 470, 60, 10, 0xFF, 0xFF, 0xFF, 300);

    	// Ball box.
    	AddBox(pMemPosition, iBallXPos, iBallYPos, 4, 4, 0xFF, 0xFF, 0xFF, 300);

		/*
		We now have written all 85 avaialble boxes (4096 bytes).
		In stead of writing these to the FIFO registers of the GIF, we wrote these to memory.
		So we now need to translafer these to the GIF.
		According to Page 42 of the EEUSER_E.PDF the GIF has channel ID 2.
		It only have one direction so we don't have to worry about setting any direction flags.
		The actual DMAC registers can be found @ 0x1000_8000 to 0x1000_EFF0.
		The second DMA channel is located @ 0x1000_A000 to 0x1000_AFF0
		*/

	  	// Set the channel control register values.
  		chcr.i32 = 0;                  // Clear everything
	  	chcr.s.MOD = CHCR_MOD_NORMAL;  // Normal mode transfer
  		chcr.s.STR = 1;                // STR=1 to start transfer

    	// Set the QWC register.
		// Every Qword is 128 bits, we are writing 256*(128/8) = 4096 bytes...
		// *EE_D2_QWC = 0x1000_A020.
    	*EE_D2_QWC=256;

    	// Set the memory address register with the start of allocated memroy.
		// Make sure we get the correct memory pos by using sps2GetPhysicalAddress in stead of the pointer.
		// *EE_D2_MADR = 0x1000_A010.
    	*EE_D2_MADR=sps2GetPhysicalAddress(pMemory->pvStart, pMemory);

    	// Start the DMA transfer.
		// Make sure we set the two previous registers first or we might transfer an unknown amount of garbage to the GIF.
		// *EE_D2_CHCR = 0x1000_A000.
    	*EE_D2_CHCR=chcr.i32;

    	// Wait for the DMA transfer to finish.
    	sps2WaitForDMA(2, iSPS2Descriptor);

    	// Swap display buffers
    	sps2UScreenSwap();

	// Press start and select together to end the loop	
	}while(!((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)));
	
	// Release the pad device and clean up.
	pad_cleanup (PAD_0);
  
	// Shut down the screen
	sps2UScreenShutdown();

	// Release the allocated memory.
	sps2Free(pMemory);

	// Release sps2
	sps2Release(iSPS2Descriptor);
  
  return 0;
}

// Just add the GIF Tag to the beginning of the allocated Memory.
void AddGIFTag(char *& pMem) {

		// Add a GIF Tag 
		sps2GIFTag_t * pGIFTag = (sps2GIFTag_t *) pMem;
 
	  	// Create the GIFTag  
  		pGIFTag->i128=0;       			// Blank it out.
	  	pGIFTag->s.NLOOP=85;   			// We are rendering 85 sprites.
  		pGIFTag->s.PRE=1;      			// Send our PRIM value to the GIF via this GIFTag.
	  	pGIFTag->s.EOP=1;      			// This is the end of the packets, no more GIFTags are to be expected after the sprites.
  		pGIFTag->s.PRIM=6;    			// Sprite, basically just a square.
	  	pGIFTag->s.FLG=GIF_FLG_PACKED;	// Use packed format for data.
  		pGIFTag->s.NREG=0;     			// Start at 0 and increment as needed.
  
	  	// Add RGBAQ register.
  		SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_RGBAQ);
	  	// Add XYZ2 register.
  		SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_XYZ2);
	  	// Add XYZ2 register.
  		SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_XYZ2);
		pGIFTag++;

		// Update memory pointer.
		pMem = (char *)pGIFTag;
}

// Add the boxes to the allocated Memory. Every box will have three GIF Regs.
void AddBox(char *& pMem, int iX, int iY, int iWidth, int iHeight, int iRed, int iGreen, int iBlue, int iZvalue)
{
	// pGIFReg is now pointing to the first memory position.
  	sps2GIFPackedRegister_t * pGIFReg = (sps2GIFPackedRegister_t *) pMem;

   	// Prepare the RGBAQ register
   	pGIFReg->i128=0;
   	pGIFReg->RGBAQ.s.R=iRed;	// Red
   	pGIFReg->RGBAQ.s.G=iGreen;	// Green
   	pGIFReg->RGBAQ.s.B=iBlue;	// Blue
   	pGIFReg->RGBAQ.s.A=0x80;	// Alpha
	pGIFReg++;

   	// Prepare the first XYZ2 register
   	pGIFReg->i128=0;
   	pGIFReg->XYZ2.s.X=(iX + 1728) << 4;
   	pGIFReg->XYZ2.s.Y=(iY + 1808) << 4;
   	pGIFReg->XYZ2.s.Z=iZvalue;
	pGIFReg++;

   	// Prepare the second XYZ2 register
   	pGIFReg->i128=0;
   	pGIFReg->XYZ2.s.X=(iX + iWidth + 1728) << 4;
   	pGIFReg->XYZ2.s.Y=(iY + iHeight + 1808) << 4;
   	pGIFReg->XYZ2.s.Z=iZvalue;;
	pGIFReg++;

	// Update memory pointer.
	pMem = (char *)pGIFReg;

}

// How many bricks are left in the game???
int CountBricks(void) {
	int iTotal = 0;
	for (int iRows = 0; iRows<4 ; iRows++) {
		for (int iCols = 0; iCols<20 ; iCols++) {
			if (iBricks[iRows][iCols]) {
				iTotal++;
			}
		}
	}
	return iTotal;
}

// Make all bricks visible.
void ResetBricks(void) {
	for (int iRows = 0; iRows<4 ; iRows++) {
		for (int iCols = 0; iCols<20 ; iCols++) {
			iBricks[iRows][iCols] = 1;
		}
	}
}
